/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  T z C L O C K C A I R O. C                                                                        *
 *  ==========================                                                                        *
 *                                                                                                    *
 *  TzClock developed by Chris Knight based on glock by Eric L. Smith.                                * 
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  TzClock is free software; you can redistribute it and/or modify it under the terms of the GNU     *
 *  General Public License version 2 as published by the Free Software Foundation.  Note that I       *
 *  am not granting permission to redistribute or modify TzClock under the terms of any later         *
 *  version of the General Public License.                                                            *
 *                                                                                                    *
 *  TzClock is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without      *
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *  GNU General Public License for more details.                                                      *
 *                                                                                                    *
 *  You should have received a copy of the GNU General Public License along with this program (in     * 
 *  the file "COPYING"); if not, write to the Free Software Foundation, Inc.,                         *
 *  59 Temple Place - Suite 330, Boston, MA 02111, USA.                                               *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @file
 *  @brief .
 */
 
#include "config.h"
#include "TzClockDisp.h"
#include "TimeZone.h"

extern GtkWidget *mainWindow;
extern GtkWidget *drawingArea;
extern GdkColor clockColours[];
extern char fontName[];
extern int faceSize;
extern int faceWidth;
extern int faceHeight;
extern int weHaveFocus;
extern int currentFace;
extern int timeSetting;
extern int fastSetting;
extern int showBounceSec;
extern int markerType;
extern int markerStep;
extern time_t forceTime;
extern HAND_STYLE handStyle[];
extern FACE_SETTINGS faceSettings[];
extern TZ_INFO *timeZones;

static GdkDrawable *windowShapeBitmap;
static char windowTitle[128];

/*------------------------------------------------------------------------------------------------*
 * Change the font size as the window gets bigger.                                                *
 *------------------------------------------------------------------------------------------------*/
static int fontSizes[16] = 
{
	 6, 8, 12, 16, 20, 24, 28, 32, 36, 40, 44, 48, 52, 56, 60, 64,
};
char *roman[24] = 
{
	"XXIV", "I", "II", "III", "IV", "V", "VI", "VII", "VIII", "IX", "X", "XI", "XII",
	"XIII", "XIV", "XV", "XVI", "XVII", "XVIII", "XIX", "XX", "XXI", "XXII", "XXIII"
};
char *latin[24] =
{
	"24", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12", 
	"13", "14", "15", "16", "17", "18", "19", "20", "21", "22", "23"
};

void drawText (cairo_t *cr, int posX, int posY, char *string1, int colour, int scale);
int getFontSize (char *fontName);

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  M A K E  W I N D O W  M A S K                                                                     *
 *  =============================                                                                     *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @result .
 */
void makeWindowMask ()
{
	int fullHeight, fullWidth, i, j;
	GdkGC *gc;
        
	fullHeight = faceHeight * faceSize;
	fullWidth = faceWidth * faceSize;
        
	if (windowShapeBitmap)
	{
		g_object_unref (windowShapeBitmap);
		windowShapeBitmap = NULL;
	}

	windowShapeBitmap = (GdkBitmap *) gdk_pixmap_new (NULL, fullWidth, fullHeight, 1);

	gc = gdk_gc_new (windowShapeBitmap);
	gdk_gc_set_foreground (gc, &clockColours[BLACK_COLOUR]);
	gdk_gc_set_background (gc, &clockColours[WHITE_COLOUR]);

	gdk_draw_rectangle (windowShapeBitmap, gc, TRUE, 0, 0, fullWidth, fullHeight);

	gdk_gc_set_foreground (gc, &clockColours[WHITE_COLOUR]);
	gdk_gc_set_background (gc, &clockColours[BLACK_COLOUR]);

	for (i = 0; i < faceWidth; i++)
	{
		for (j = 0; j < faceHeight; j++)
			gdk_draw_arc (windowShapeBitmap, gc, TRUE, (i * faceSize) - 1, (j * faceSize) - 1, 
					faceSize + 1, faceSize + 1, 0, 360 * 64);
	}
	gtk_widget_shape_combine_mask (mainWindow, windowShapeBitmap, 0, 0);
        
	g_object_unref (gc);
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  D R A W  M I N U T E                                                                              *
 *  ====================                                                                              *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param cr .
 *  @param posX .
 *  @param posY .
 *  @param size .
 *  @param len .
 *  @param angle .
 *  @param colour .
 *  @result .
 */
void drawMinute (cairo_t *cr, int posX, int posY, int size, int len, int angle, int colour)
{
	gdk_cairo_set_source_color (cr, &clockColours[colour]);
	if (len < size)
	{
		cairo_move_to (cr,
				posX + xSinCos ((faceSize * size) >> 6, angle, 0),
				posY - xSinCos ((faceSize * size) >> 6, angle, 1));

		cairo_line_to (cr,
				posX + xSinCos ((faceSize * (size + len)) >> 6, angle, 0),
				posY - xSinCos ((faceSize * (size + len)) >> 6, angle, 1));
	}
	else
	{
		cairo_move_to (cr,
				posX + xSinCos ((faceSize * size) >> 6, angle + 180, 0),
				posY - xSinCos ((faceSize * size) >> 6, angle + 180, 1));

		cairo_line_to (cr,
				posX + xSinCos ((faceSize * (size + len)) >> 6, angle, 0),
				posY - xSinCos ((faceSize * (size + len)) >> 6, angle, 1));
	}
	cairo_stroke (cr);
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  D R A W  C I R C L E                                                                              *
 *  ====================                                                                              *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param cr .
 *  @param posX .
 *  @param posY .
 *  @param size .
 *  @param colFill .
 *  @param colOut .
 *  @result .
 */
void drawCircle (cairo_t *cr, int posX, int posY, int size, int colFill, int colOut)
{
	if (colFill != -1)
	{
		gdk_cairo_set_source_color (cr, &clockColours[colFill]);
 		cairo_arc (cr, posX, posY, (faceSize * size) >> 7, 0, 2 * M_PI);
		cairo_fill (cr);
		cairo_stroke (cr);
	}
	if (colOut != -1)
	{
		gdk_cairo_set_source_color (cr, &clockColours[colOut]);
 		cairo_arc (cr, posX, posY, (faceSize * size) >> 7, 0, 2 * M_PI);
		cairo_stroke (cr);
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  D R A W  H A N D                                                                                  *
 *  ================                                                                                  *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param cr .
 *  @param posX .
 *  @param posY .
 *  @param size .
 *  @param angle .
 *  @param colFill .
 *  @param colOut .
 *  @result .
 */
void drawHand (cairo_t *cr, int posX, int posY, int angle, HAND_STYLE *handStyle)
{
	int points[20], i, j, numPoints, fill = 0;
	int size = handStyle -> length, style = handStyle -> style, tail = handStyle -> tail;
	int colFill = handStyle -> fill, colOut = handStyle -> line;

	switch (style)
	{
	case 0:
		// Original double triangle
		points[0] = posX + xSinCos ((faceSize * tail) >> 6, angle + 180, 0);
		points[1] = posY - xSinCos ((faceSize * tail) >> 6, angle + 180, 1);
		points[2] = posX + xSinCos (faceSize >> 6, angle + 90, 0);
		points[3] = posY - xSinCos (faceSize >> 6, angle + 90, 1);
		points[4] = posX + xSinCos ((faceSize * size) >> 6, angle, 0);
		points[5] = posY - xSinCos ((faceSize * size) >> 6, angle, 1);
		points[6] = posX + xSinCos (faceSize >> 6, angle + 270, 0);
		points[7] = posY - xSinCos (faceSize >> 6, angle + 270, 1); 
		numPoints = 4;
		fill = 1;
		break;

	case 1:
		// Single triangle
		points[0] = posX + xSinCos ((faceSize * tail) >> 6, angle + 180, 0) 
					+ xSinCos (faceSize >> 6, angle + 90, 0);
		points[1] = posY - xSinCos ((faceSize * tail) >> 6, angle + 180, 1)
					- xSinCos (faceSize >> 6, angle + 90, 1);
		points[2] = posX + xSinCos ((faceSize * tail) >> 6, angle + 180, 0) 
					+ xSinCos (faceSize >> 6, angle + 270, 0);
		points[3] = posY - xSinCos ((faceSize * tail) >> 6, angle + 180, 1)
					- xSinCos (faceSize >> 6, angle + 270, 1);
		points[4] = posX + xSinCos ((faceSize * size) >> 6, angle, 0);
		points[5] = posY - xSinCos ((faceSize * size) >> 6, angle, 1);
		numPoints = 3;
		fill = 1;
		break;

	case 2:
		// Rectangle
		points[0] = posX + xSinCos ((faceSize * tail) >> 6, angle + 180, 0) + xSinCos (faceSize >> 6, angle + 90, 0);
		points[1] = posY - xSinCos ((faceSize * tail) >> 6, angle + 180, 1) - xSinCos (faceSize >> 6, angle + 90, 1);
		points[2] = posX + xSinCos ((faceSize * tail) >> 6, angle + 180, 0) + xSinCos (faceSize >> 6, angle + 270, 0);
		points[3] = posY - xSinCos ((faceSize * tail) >> 6, angle + 180, 1) - xSinCos (faceSize >> 6, angle + 270, 1);
		points[4] = posX + xSinCos ((faceSize * size) >> 6, angle, 0) + xSinCos (faceSize >> 6, angle + 270, 0);
		points[5] = posY - xSinCos ((faceSize * size) >> 6, angle, 1) - xSinCos (faceSize >> 6, angle + 270, 1);
		points[6] = posX + xSinCos ((faceSize * size) >> 6, angle, 0) + xSinCos (faceSize >> 6, angle + 90, 0);
		points[7] = posY - xSinCos ((faceSize * size) >> 6, angle, 1) - xSinCos (faceSize >> 6, angle + 90, 1);
		numPoints = 4;
		fill = 1;
		break;

	case 3:
		//Rectangle with pointer
		points[0] = posX + xSinCos ((faceSize * tail) >> 6, angle + 180, 0) + xSinCos (faceSize >> 6, angle + 90, 0);
		points[1] = posY - xSinCos ((faceSize * tail) >> 6, angle + 180, 1) - xSinCos (faceSize >> 6, angle + 90, 1);
		points[2] = posX + xSinCos ((faceSize * tail) >> 6, angle + 180, 0) + xSinCos (faceSize >> 6, angle + 270, 0);
		points[3] = posY - xSinCos ((faceSize * tail) >> 6, angle + 180, 1) - xSinCos (faceSize >> 6, angle + 270, 1);
		points[4] = posX + xSinCos ((faceSize * (size * 15)) >> 10, angle, 0) + xSinCos (faceSize >> 6, angle + 270, 0);
		points[5] = posY - xSinCos ((faceSize * (size * 15)) >> 10, angle, 1) - xSinCos (faceSize >> 6, angle + 270, 1);
		points[6] = posX + xSinCos ((faceSize * size) >> 6, angle, 0);
		points[7] = posY - xSinCos ((faceSize * size) >> 6, angle, 1);
		points[8] = posX + xSinCos ((faceSize * (size * 15)) >> 10, angle, 0) + xSinCos (faceSize >> 6, angle + 90, 0);
		points[9] = posY - xSinCos ((faceSize * (size * 15)) >> 10, angle, 1) - xSinCos (faceSize >> 6, angle + 90, 1);
		numPoints = 5;
		fill = 1;
		break;

	case 4:
		//Rectangle with arrow
		points[0] = posX + xSinCos ((faceSize * tail) >> 6, angle + 180, 0) + xSinCos (faceSize >> 6, angle + 90, 0);
		points[1] = posY - xSinCos ((faceSize * tail) >> 6, angle + 180, 1)	- xSinCos (faceSize >> 6, angle + 90, 1);
		points[2] = posX + xSinCos ((faceSize * tail) >> 6, angle + 180, 0)	+ xSinCos (faceSize >> 6, angle + 270, 0);
		points[3] = posY - xSinCos ((faceSize * tail) >> 6, angle + 180, 1)	- xSinCos (faceSize >> 6, angle + 270, 1);
		points[4] = posX + xSinCos ((faceSize * (size * 12)) >> 10, angle, 0) + xSinCos (faceSize >> 6, angle + 270, 0);
		points[5] = posY - xSinCos ((faceSize * (size * 12)) >> 10, angle, 1) - xSinCos (faceSize >> 6, angle + 270, 1);
		points[6] = posX + xSinCos ((faceSize * (size * 12)) >> 10, angle, 0) + xSinCos (faceSize / 30, angle + 270, 0);
		points[7] = posY - xSinCos ((faceSize * (size * 12)) >> 10, angle, 1) - xSinCos (faceSize / 30, angle + 270, 1);
		points[8] = posX + xSinCos ((faceSize * size) >> 6, angle, 0);
		points[9] = posY - xSinCos ((faceSize * size) >> 6, angle, 1);
		points[10] = posX + xSinCos ((faceSize * (size * 12)) >> 10, angle, 0) + xSinCos (faceSize / 30, angle + 90, 0);
		points[11] = posY - xSinCos ((faceSize * (size * 12)) >> 10, angle, 1) - xSinCos (faceSize / 30, angle + 90, 1);
		points[12] = posX + xSinCos ((faceSize * (size * 12)) >> 10, angle, 0) + xSinCos (faceSize >> 6, angle + 90, 0);
		points[13] = posY - xSinCos ((faceSize * (size * 12)) >> 10, angle, 1) - xSinCos (faceSize >> 6, angle + 90, 1);
		numPoints = 7;
		fill = 1;
		break;

	case 9:
	default:
		// Simple line
		points[0] = posX + xSinCos ((faceSize * tail) >> 6, angle + 180, 0);
		points[1] = posY - xSinCos ((faceSize * tail) >> 6, angle + 180, 1);
		points[2] = posX + xSinCos ((faceSize * size) >> 6, angle, 0);
		points[3] = posY - xSinCos ((faceSize * size) >> 6, angle, 1);
		numPoints = 2;
		break;
	}
	if (!handStyle -> fillIn)
	{
		fill = 0;
	}
	for (i = 0; i < 2; i++)
	{
		if (i != 0 || fill)
		{
			gdk_cairo_set_source_color (cr, &clockColours[i == 0 ? colFill : colOut]);
			cairo_move_to (cr, points[0], points[1]);
			for (j = 1; j < numPoints; j++)
				cairo_line_to (cr, points[j << 1], points[(j << 1) + 1]);
			if (numPoints > 2)
				cairo_close_path (cr);
			if (i == 0)
				cairo_fill (cr);
			cairo_stroke (cr);
		}
	}
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  D R A W  M A R K                                                                                  *
 *  ================                                                                                  *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param cr .
 *  @param posX .
 *  @param posY .
 *  @param angle .
 *  @param size .
 *  @param colFill .
 *  @param colOut .
 *  @result .
 */
void drawMark (cairo_t *cr, int posX, int posY, int angle, int size, 
		int colFill, int colOut, int hour)
{
	if (markerStep == 0) markerStep = 1;

	if (!(hour % markerStep))
	{
		switch (markerType)
		{
		case 0:
			/* No markers */
			break;
	
		case 1:
			/* Triangle markers */
			{
				int points[6], i;
				points[0] = posX + xSinCos ((faceSize * size) >> 6, angle + 2, 0);
				points[1] = posY - xSinCos ((faceSize * size) >> 6, angle + 2, 1);
				points[2] = posX + xSinCos ((faceSize * (size - 3)) >> 6, angle, 0);
				points[3] = posY - xSinCos ((faceSize * (size - 3)) >> 6, angle, 1);
				points[4] = posX + xSinCos ((faceSize * size) >> 6, angle - 2, 0);
				points[5] = posY - xSinCos ((faceSize * size) >> 6, angle - 2, 1);
				for (i = 0; i < 2; i++)
				{
					gdk_cairo_set_source_color (cr, &clockColours[i == 0 ? colFill : colOut]);
					cairo_move_to (cr, points[0], points[1]);
					cairo_line_to (cr, points[2], points[3]);
					cairo_line_to (cr, points[4], points[5]);
					cairo_close_path (cr);
					if (i == 0)	cairo_fill (cr);
					cairo_stroke (cr);
				}
			}
			break;

		case 2:
			/* Circle markers */
			posX += xSinCos ((faceSize * (size - 2)) >> 6, angle, 0);
			posY -= xSinCos ((faceSize * (size - 2)) >> 6, angle, 1);
			drawCircle (cr, posX, posY, 3, colFill, colOut);
			break;

		case 3:
			/* Latin number markers */
			posX += xSinCos ((faceSize * (size - 5)) >> 6, angle, 0);
			posY -= xSinCos ((faceSize * (size - 5)) >> 6, angle, 1);
			drawText (cr, posX, posY, latin[hour], colOut, 1);
			break;

		case 4:
			/* Roman number markers */
			posX += xSinCos ((faceSize * (size - 5)) >> 6, angle, 0);
			posY -= xSinCos ((faceSize * (size - 5)) >> 6, angle, 1);
			drawText (cr, posX, posY, roman[hour], colOut, 1);
			break;
		}
	}
}

int getFontSize (char *fontName)
{
	int size = 0, i = strlen (fontName), n = 1;
	
	while (i)
	{
		i --;
		if (fontName[i] >= '0' && fontName[i] <= '9')
		{
			size += (fontName[i] - '0') * n;
			n *= 10;
		}
		else
			break;
	}
	return size;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  D R A W  T E X T                                                                                  *
 *  ================                                                                                  *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param cr .
 *  @param posX .
 *  @param posY .
 *  @param string1 .
 *  @param string2 .
 *  @result .
 */
void drawText (cairo_t *cr, int posX, int posY, char *string1, int colour, int scale)
{
	int fontSize = 0;
	gint posW, posH;

	if ((fontSize = getFontSize (fontName)) == 0)
		fontSize = fontSizes[(faceSize >> 6) - 1];
	if (scale)
		fontSize = (fontSize * 10) / 12;

	PangoLayout *layout = pango_cairo_create_layout (cr);;
	PangoFontDescription *fontDesc = pango_font_description_from_string (fontName);

	pango_font_description_set_size (fontDesc, PANGO_SCALE * fontSize);
 	pango_layout_set_font_description (layout, fontDesc); 
	pango_layout_set_alignment (layout, PANGO_ALIGN_CENTER);

	gdk_cairo_set_source_color (cr, &clockColours[colour == -1 ? TEXT__COLOUR : colour]);
	pango_layout_set_text (layout, string1, -1);
	pango_layout_get_pixel_size (layout, &posW, &posH);

	cairo_move_to (cr, posX - (posW >> 1), posY - (posH >> 1));
	pango_cairo_show_layout (cr, layout);

	pango_font_description_free (fontDesc);
	g_object_unref (layout);
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  D R A W  F A C E                                                                                  *
 *  ================                                                                                  *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param cr .
 *  @param face .
 *  @param posX .
 *  @param posY .
 *  @param t .
 *  @result .
 */
gboolean 
drawFace (cairo_t *cr, int face, int posX, int posY, time_t t, int bounce)
{
	int i, j, tempTime, timeZone = faceSettings[face].currentTZ;
	int hourAngle, minuteAngle, secondAngle, swMinAngle, swSecAngle = 0, swPrtAngle;
	int centerX = posX + (faceSize >> 1), centerY = posY + (faceSize >> 1);
	int showSubSec, markerFlags = 0xFFFFFF;
	char tempString[101];
	struct tm tm;

	/*------------------------------------------------------------------------------------------------*
	 * The alarm check has moved out of here                                                          *
	 *------------------------------------------------------------------------------------------------*/	
	getTheFaceTime (face, t, &tm);
	
	/*------------------------------------------------------------------------------------------------*
	 * Handle the smooth scroll time change                                                           *
	 *------------------------------------------------------------------------------------------------*/	
	tempTime = (tm.tm_hour * 60) + tm.tm_min;
		
	if (!fastSetting)
	{
		int gap = 0;
		
		/*--------------------------------------------------------------------------------------------*
		 * Don't scroll at midnight                                                                   *
		 *--------------------------------------------------------------------------------------------*/
		if (tempTime == 0 && faceSettings[face].shownTime == (23 * 60) + 59)
			gap = 1;
		else
			gap = tempTime - faceSettings[face].shownTime;
		
		if (abs (gap) > 1)
		{		
			/*----------------------------------------------------------------------------------------*
			 * The maximum gap is 12 hours unless it is a 24 hour clock                               *
			 *----------------------------------------------------------------------------------------*/
			if (abs(gap) > TWELVEHOURS && !faceSettings[face].show24Hour)
			{
				if (gap > 0)
				{
					faceSettings[face].shownTime += TWELVEHOURS;
					gap -= TWELVEHOURS;
				}
				else
				{
					faceSettings[face].shownTime -= TWELVEHOURS;
					gap += TWELVEHOURS;
				}
			}

			j = 1;
			for (i = 1; i <= faceSettings[face].stepping; i ++)
			{
				j += (i / 3);
				if (j > abs (gap))
					break;
			}
			i /= 3;
			if (gap < 0)
				i *= -1;

    	    if (faceSettings[face].stepping < 24)
				faceSettings[face].stepping ++;

			tempTime = faceSettings[face].shownTime + i;
			
			tm.tm_min = tempTime % 60;
			tm.tm_hour = tempTime / 60;
			
			timeSetting = 1;
		}
		else
			faceSettings[face].stepping = 0;
	}
	faceSettings[face].shownTime = tempTime;

	/*------------------------------------------------------------------------------------------------*
	 * Draw the face, it is made up of 3 overlapping circles                                          *
	 *------------------------------------------------------------------------------------------------*/	
	cairo_save (cr);
	cairo_set_line_width (cr, 1.0f + ((float)faceSize / 256.0f));
	cairo_set_line_cap (cr, CAIRO_LINE_CAP_BUTT);
	cairo_set_line_join (cr, CAIRO_LINE_JOIN_MITER);

	if (weHaveFocus && face == currentFace)
		gdk_cairo_set_source_color (cr, &clockColours[FACE1_COLOUR]);
	else
		gdk_cairo_set_source_color (cr, &clockColours[FACE2_COLOUR]);

	cairo_rectangle (cr, posX, posY, faceSize, faceSize);
	cairo_fill (cr);
	drawCircle (cr, centerX, centerY, 62, FACE3_COLOUR, -1);
	drawCircle (cr, centerX, centerY, 60, FACE4_COLOUR, -1);
	
	/*------------------------------------------------------------------------------------------------*
	 * Add the text, ether the date or the timezone, plus an AM/PM indicator                          *
	 *------------------------------------------------------------------------------------------------*/
	showSubSec = (faceSettings[face].showSeconds && (faceSettings[face].stopwatch || faceSettings[face].subSecond)) ? 1 : 0;

	getStringValue (tempString, NULL, 100, faceSettings[face].stopwatch ?
			(timeZone ? TXT_TOPSW_Z : TXT_TOPSW_L) : (timeZone ? TXT_TOP_Z : TXT_TOP_L), face, t);
	drawText (cr, centerX, posY + ((faceSize * 5) >> 4), tempString, -1, 0);
	 
	if (!showSubSec)
	{
		getStringValue (tempString, NULL, 100, timeZone ? TXT_BOTTOM_Z : TXT_BOTTOM_L, face, t);	
		drawText (cr, centerX, posY + ((faceSize * 11) >> 4), tempString, -1, 0);
	}

	if (face == currentFace)
	{
		getStringValue (tempString, NULL, 100, timeZone ? TXT_TITLE_Z : TXT_TITLE_L, face, t);
		if (strcmp (windowTitle, tempString))
		{
			strcpy (windowTitle, tempString);
			gtk_window_set_title (GTK_WINDOW (mainWindow), windowTitle);
		}
	}

	/*------------------------------------------------------------------------------------------------*
	 * Calculate which markers to draw                                                                *
	 *------------------------------------------------------------------------------------------------*/
	if (markerType > 2)
	{
		if (showSubSec)
		{
			if (faceSettings[face].show24Hour)
			{
				markerFlags &= ~(1 << 11);
				markerFlags &= ~(1 << 12);
				markerFlags &= ~(1 << 13);
			}
			else
				markerFlags &= ~(1 << 6);
		}
		if (faceSettings[face].stopwatch)
		{
			if (faceSettings[face].show24Hour)
			{
				markerFlags &= ~(1 << 5);
				markerFlags &= ~(1 << 6);
				markerFlags &= ~(1 << 7);
				markerFlags &= ~(1 << 17);
				markerFlags &= ~(1 << 18);
				markerFlags &= ~(1 << 19);
			}
			else
			{
				markerFlags &= ~(1 << 3);
				markerFlags &= ~(1 << 9);
			}
		}		
	}

	/*------------------------------------------------------------------------------------------------*
	 * Draw the hour markers                                                                          *
	 *------------------------------------------------------------------------------------------------*/
	cairo_set_line_width (cr, 1.0f + ((float)faceSize / 512.0f));
	j = faceSettings[face].show24Hour ? 120 : 60;

	for (i = 0; i < j ; i++)
	{
		hourAngle = faceSettings[face].show24Hour ? i * 3 : i * 6;

		if (faceSize > 256)
		{
			if (!faceSettings[face].show24Hour || !(i % 2))
				drawMinute (cr, centerX, centerY, 30, 1, hourAngle, MMARK_COLOUR);
		}
		if (!(i % 5))
		{
			drawMinute (cr, centerX, centerY, 29, 1, hourAngle, HMARK_COLOUR);
			if (markerFlags & (1 << (i / 5)))
				drawMark (cr, centerX , centerY, hourAngle, 31, QFILL_COLOUR, QMARK_COLOUR, 
						(i == 0 & !faceSettings[face].show24Hour) ? 12 : i / 5);
		}
	}

	/*------------------------------------------------------------------------------------------------*
	 * Draw other clock faces                                                                         *
	 *------------------------------------------------------------------------------------------------*/
	if (showSubSec)
	{
		drawCircle (cr, centerX, posY + ((3 * faceSize) >> 2), 21, FACE5_COLOUR, -1);
	}
	if (faceSettings[face].stopwatch)
	{
		drawCircle (cr, posX + (faceSize >> 2), centerY, 21, FACE5_COLOUR, -1);			
		drawCircle (cr, posX + (3 * faceSize >> 2), centerY, 21, FACE5_COLOUR, -1);
	}	
	
	for (i = 0; i < 60 ; i++)
	{
		hourAngle = i * 6;

		if (showSubSec)
		{
			if (!(i % 5))
				drawMinute (cr, centerX, posY + ((3 * faceSize) >> 2),
						(i % 15) ? 9 : 8, (i % 15) ? 1 : 2, hourAngle, WMARK_COLOUR);
		}
		if (faceSettings[face].stopwatch)
		{
			if (!(i % 6))
				drawMinute (cr, posX + (faceSize >> 2), centerY,
						(i % 12) ? 9 : 8, (i % 12) ? 1 : 2, hourAngle, WMARK_COLOUR);
			if (!(i % 4))
				drawMinute (cr, posX + ((3 * faceSize) >> 2), centerY,
						(i % 20) ? 9 : 8, (i % 20) ? 1 : 2, hourAngle, WMARK_COLOUR);
		}
	}

	cairo_set_line_width (cr, 1.0f + ((float)faceSize / 256.0f));

	/*------------------------------------------------------------------------------------------------*
	 * Draw the hands                                                                                 *
	 *------------------------------------------------------------------------------------------------*/
	secondAngle = (tm.tm_sec * 6);
	minuteAngle = (tm.tm_min * 6) + (tm.tm_sec / 10);
	if (faceSettings[face].show24Hour)	
		hourAngle = (tm.tm_hour * 15) + (tm.tm_min >> 2);
	else
		hourAngle = (tm.tm_hour * 30) + (tm.tm_min >> 1);

	if (faceSettings[face].alarmInfo.showAlarm)
	{
		int alarmAngle;
		
		if (faceSettings[face].show24Hour)	
			alarmAngle = (faceSettings[face].alarmInfo.alarmHour * 15) + (faceSettings[face].alarmInfo.alarmMin >> 2);
		else
			alarmAngle = (faceSettings[face].alarmInfo.alarmHour * 30) + (faceSettings[face].alarmInfo.alarmMin >> 1);
	
		drawHand (cr, centerX, centerY, alarmAngle, &handStyle[HAND_ALARM]);
	}
	if (showSubSec)
	{
		drawHand (cr, centerX, posY + ((3 * faceSize) >> 2), secondAngle, &handStyle[HAND_SUBS]);
		drawCircle (cr, centerX, posY + ((3 * faceSize) >> 2), 2, SFILL_COLOUR, SEC___COLOUR);
	}
	if (faceSettings[face].stopwatch)
	{
		int swTime = getStopwatchTime(face);
		swMinAngle = (swTime / 6000) * 12;
		swSecAngle = ((swTime / 100) % 60) * 6;
		swPrtAngle = ((swTime % 100) * 36) / 10;

		drawHand (cr, posX + (faceSize >> 2), centerY, swPrtAngle, &handStyle[HAND_STOPW]);
		drawHand (cr, posX + ((3 * faceSize) >> 2), centerY, swMinAngle, &handStyle[HAND_STOPW]);
		drawCircle (cr, posX + (faceSize >> 2), centerY, 2, WFILL_COLOUR, WATCH_COLOUR);			
		drawCircle (cr, posX + ((3 * faceSize) >> 2), centerY, 2, WFILL_COLOUR, WATCH_COLOUR);
	}

	drawHand (cr, centerX, centerY, hourAngle, &handStyle[HAND_HOUR]);
	drawHand (cr, centerX, centerY, minuteAngle, &handStyle[HAND_MINUTE]);

	if (faceSettings[face].showSeconds || faceSettings[face].stopwatch)
	{
		if (faceSettings[face].stopwatch)
		{
			drawHand (cr, centerX, centerY, swSecAngle, &handStyle[HAND_SECS]);
			drawCircle (cr, centerX, centerY, 4, WFILL_COLOUR, WATCH_COLOUR);
		}
		else if (faceSettings[face].subSecond)
		{
			drawCircle (cr, centerX, centerY, 4, MFILL_COLOUR, MIN___COLOUR);
		}
		else
		{
			secondAngle += bounce;
			drawHand (cr, centerX, centerY, secondAngle, &handStyle[HAND_SECS]);
			drawCircle (cr, centerX, centerY, 4, SFILL_COLOUR, SEC___COLOUR);
		}
	}
	else
	{
		drawCircle (cr, centerX, centerY, 4, MFILL_COLOUR, MIN___COLOUR);
	}
	cairo_restore (cr);
	return TRUE;
}

/*----------------------------------------------------------------------------------------------------*
 *                                                                                                    *
 *  C L O C K  E X P O S E                                                                            *
 *  ======================                                                                            *
 *                                                                                                    *
 *----------------------------------------------------------------------------------------------------*/
/**
 *  @brief .
 *  @param widget .
 *  @result .
 */
void clockExpose (GtkWidget * widget)
{
	int i, j, face = 0, bounce = 0;
	cairo_t *cr = gdk_cairo_create (drawingArea -> window);
	struct timeval tv;
	struct timezone tz;
		
	gettimeofday(&tv, &tz);
	if (showBounceSec)
		bounce = tv.tv_usec < 50000 ? 2 : 0;

	/*------------------------------------------------------------------------------------------------*
	 * Reset the color and stuff back to the default.                                                 *
	 *------------------------------------------------------------------------------------------------*/
	timeSetting = 0;

	if (forceTime != -1)
	{
		tv.tv_sec = forceTime;
		bounce = 0;
	}
		
	for (j = 0; j < faceHeight; j++)
	{
		for (i = 0; i < faceWidth; i++)
			drawFace (cr, face++, (i * faceSize), (j * faceSize), tv.tv_sec, bounce);
	}

	/*------------------------------------------------------------------------------------------------*
	 * Reset the color and stuff back to the default.                                                 *
	 *------------------------------------------------------------------------------------------------*/
	cairo_destroy (cr);
	cr = NULL;	
}

